/*
   Copyright 2006-2009 Murali Krishnan

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */

package com.notuvy.thread;

import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.atomic.AtomicLong;
import java.io.InputStreamReader;
import java.io.BufferedReader;

import org.apache.log4j.Logger;
import com.notuvy.util.WallClock;

/**
 * Basically an infinite loop with a sleep period (to avoid CPU
 * saturation).
 * It is used to continuously execute an action (Runnable) in a
 * separate Thread.
 * Once started, it can be suspended (disabled), resumed (enabled),
 * and stopped.
 *
 * @author  murali
 */
public class ContinuousExecution implements ThreadControllable {

    //------------------------------------------------------------
    //- Class Variables

    protected static final Logger LOG = Logger.getLogger(ContinuousExecution.class);


    //------------------------------------------------------------
    //- Class Functions


    //------------------------------------------------------------
    //- Instance Variables

    private final ThreadControl fControl = new ThreadControl();
    private long fFireTime = 0;
    private final AtomicLong fIterations = new AtomicLong();
    private int fPeriod = 1000;
    private Runnable fAction;
    private ScheduledExecutorService fService;
    private ScheduledFuture fFuture = null;
    private NamedThreadFactory fThreadFactory = null;


    //------------------------------------------------------------
    //- Constructors

    @Deprecated
    public ContinuousExecution(int pPeriod) {
        setPeriod(pPeriod);
    }

    @Deprecated
    public ContinuousExecution(long pPeriod) {
        setPeriod((int) pPeriod);
    }

    public ContinuousExecution(int pPeriod, Runnable pRunnable, String pThreadName) {
        setPeriod(pPeriod);
        fAction = pRunnable;
        createThreadFactory(pThreadName);
    }

    public ContinuousExecution(long pPeriod, Runnable pRunnable, String pThreadName) {
        this((int)pPeriod, pRunnable, pThreadName);
    }


    //------------------------------------------------------------
    //- Accessors

    public int getPeriod() {
        return fPeriod;
    }

    public ThreadControl getControl() {
        return fControl;
    }

    private long getFireTime() {
        return fFireTime;
    }


    //------------------------------------------------------------
    //- Settors

    private void setPeriod(int pPeriod) {
        fPeriod = pPeriod;
    }

    private void setFireTime(long pFireTime) {
        fFireTime = pFireTime;
    }


    //------------------------------------------------------------
    //- Private/Protected Utility Functions

    private void createThreadFactory(String pName) {
        fThreadFactory = new NamedThreadFactory(pName);
    }

    private void schedule(long pDelay) {
        fFuture = fService.schedule(new Runner(), pDelay, TimeUnit.MILLISECONDS);
    }

    private void reschedule() {
        if (getControl().isAlive()) {
            long workTime = WallClock.currentTimeMillis() - getFireTime();
            long delay = Math.max(getPeriod() - workTime, 0);  // If the work time is greater than the wait time, then fire next immediately.
            schedule(delay);
        } else {
            synchronized (this) {
                notifyAll();  // In case anyone is waiting for this to finish.
            }
        }
    }


    //------------------------------------------------------------
    //- Public Interface Functions

    @Deprecated
    public void add(Runnable pChild) {
        fAction = pChild;
    }


    @Deprecated
    public void startThread(String pName) {
        startThread(pName, false);
    }

    @Deprecated
    public void startThread(String pName, boolean pRunImmediately) {
        if (fService == null) {
            createThreadFactory(pName);
            start(pRunImmediately);
        } else {
            LOG.error("Thread service was already started.  Ignoring latest attempt.");
        }
    }

    public void start() {
        start(false);
    }

    public void start(boolean pRunImmediately) {
        if (fService == null) {
            fService = Executors.newSingleThreadScheduledExecutor(fThreadFactory);
            int delay = pRunImmediately ? 0 : getPeriod();
            schedule(delay);
        } else {
            LOG.error("Thread service was already started.  Ignoring latest attempt.");
        }
    }

    /**
     * If the thread is sleeping, force it to run now.
     */
    public void triggerNow() {
        // First check if anything was scheduled.
        if (fFuture != null) {
            // Now try to cancel (but do not interrupt an active thread).
            // If waiting work was canceled, reschedule it to go immediately
            if (fFuture.cancel(false)) {
                schedule(0);
            }
        }
    }

    /**
     * Dynamically change the wait period.  If work is overdue (based on the new value)
     * then the work should trigger immediately.
     *
     * @param pNewPeriod  The wait period in milliseconds.
     */
    public synchronized void changePeriod(int pNewPeriod) {
        // Only act if there is an actual change.
        if (pNewPeriod != getPeriod()) {
            // Check for a valid value.
            if (pNewPeriod > 0) {
                // Check if we are in the same thread as the executor.
                if ((fThreadFactory == null)) {
                    // No thread started yet.  Just set the value.
                    setPeriod(pNewPeriod);
                } else if (fThreadFactory.originatedCurrent()) {
                    // In the executor thread.  Just set the value because reschedule() will be done.
                    setPeriod(pNewPeriod);
                } else {
                    // Different thread.  Cancel the waiting executor (if one exists) and schedule a new one.
                    long nextFire;
                    if (fFuture == null) {
                        nextFire = pNewPeriod;
                    } else {
                        long remaining = fFuture.getDelay(TimeUnit.MILLISECONDS);
                        long alreadyWaited = getPeriod() - remaining;
                        nextFire = Math.max(pNewPeriod - alreadyWaited, 0);
                        fFuture.cancel(false);
                    }

                    setPeriod(pNewPeriod);
                    schedule(nextFire);
                }
            } else {
                LOG.error(String.format("Invalid sleep time: %d ms.", pNewPeriod));
            }
        }
    }

    public synchronized void changePeriod(long pNewPeriod) {
        changePeriod((int) pNewPeriod);
    }


    @Deprecated
    /**
     * @deprecated Use getControl().suspend() instead.
     */
    public void suspend() {
        getControl().suspend();
    }

    @Deprecated
    /**
     * @deprecated Use getControl().resume() instead.
     */
    public void resume() {
        getControl().resume();
    }

    @Deprecated
    /**
     * @deprecated Use getControl().stop() instead.
     */
    public void stop() {
        getControl().stop();
    }

    public void waitForTermination() throws InterruptedException {
        synchronized (this) {
            wait();
        }
    }


    //------------------------------------------------------------
    //- Class Interface Functions

    @Override
    public String toString() {
        return String.format("ContinuousExecution(%d ms)[%d iterations]: %s", getPeriod(), fIterations.get(), getControl().toString());
    }


    //------------------------------------------------------------
    //- Inner Classes

    // Keep the run() logic in a separate class so the it is not accessible externally.
    private class Runner implements Runnable {

        public synchronized void run() {
            // The task has fired, so the future action is now running.
            fFuture = null;

            setFireTime(WallClock.currentTimeMillis());

            if (getControl().isEnabled() && getControl().isAlive()) {
                fIterations.incrementAndGet();
                try {
                    ThreadMonitor.update("Fired.");
                    getControl().setRunning(true);
                    fAction.run();
                } catch (Throwable thr) {
                    LOG.error("Fatal error.", thr);
                } finally {
                    getControl().setRunning(false);
                    ThreadMonitor.update("Finished.");
                }
            }

            reschedule();
        }

    }


    //------------------------------------------------------------
    //- Main

    public static void main(String[] args) {
        try {
            MultiRunnable indirection = new MultiRunnable();
            final ContinuousExecution runner = new ContinuousExecution(2000, indirection, "TestAction");
            indirection.add(new Runnable() {
                public void run() {
                    LOG.info(runner + ", " + WallClock.currentDate());
                    runner.changePeriod(runner.getPeriod() * 2);
                }
            });
            runner.start();

            Pattern pattern = Pattern.compile("^\\s*([0-9]+)\\s*$");
            InputStreamReader isr = new InputStreamReader(System.in);
            BufferedReader br = new BufferedReader(isr);
            String response;
            do {
                response = br.readLine();

                Matcher matcher = pattern.matcher(response);
                if (matcher.find()) {
                    int newPeriod = Integer.parseInt(matcher.group(1));
                    runner.changePeriod(newPeriod);
                } else if ("n".equals(response)) {
                    runner.triggerNow();
                }
            } while (!"q".equalsIgnoreCase(response));
            br.close();
            isr.close();

            runner.getControl().stop();
        } catch (Throwable thr) {
            LOG.error("Fatal error.", thr);
        }
    }


}
